@@ -26,7 +26,7 @@ use termion::input::Keys;
 use termion::{color, event::Key, input::TermRead};
 use textgen::{RawWordSelector, WordSelector};
 use tui::{Text, ToipeTui};
-use wordlists::{get_word_list, OS_WORDLIST_PATH};
+use wordlists::{BuiltInWordlist, OS_WORDLIST_PATH};
 
 /// Typing test terminal UI and logic.
 pub struct Toipe {
@@ -66,14 +66,18 @@ impl<'a> Toipe {
     /// Also invokes [`Toipe::restart()`].
     pub fn new(config: ToipeConfig) -> Result<Self, ToipeError> {
         let word_selector: Box<dyn WordSelector> =
-            if let Some(word_list) = get_word_list(config.wordlist.as_str()) {
+            if let Some(wordlist_path) = config.wordlist_file.clone() {
+                Box::new(RawWordSelector::from_path(PathBuf::from(wordlist_path))?)
+            } else if let Some(word_list) = config.wordlist.contents() {
                 Box::new(RawWordSelector::from_string(word_list.to_string())?)
-            } else if config.wordlist == "os" {
+            } else if let BuiltInWordlist::OS = config.wordlist {
                 Box::new(RawWordSelector::from_path(PathBuf::from(OS_WORDLIST_PATH))?)
             } else {
-                Box::new(RawWordSelector::from_path(PathBuf::from(
-                    config.wordlist.clone(),
-                ))?)
+                // this should never happen!
+                // TODO: somehow enforce this at compile time?
+                return Err(ToipeError {
+                    msg: "Undefined word list or path.".to_string(),
+                });
             };
 
         let mut toipe = Toipe {
@@ -98,6 +102,13 @@ impl<'a> Toipe {
 
         self.words = self.word_selector.new_words(self.config.num_words)?;
 
+        self.tui.display_lines_bottom(&[&[
+            Text::from("ctrl-r").with_color(color::Blue),
+            Text::from(" to restart, ").with_faint(),
+            Text::from("ctrl-c").with_color(color::Blue),
+            Text::from(" to quit ").with_faint(),
+        ]])?;
+
         self.show_words()?;
 
         Ok(())
@@ -127,19 +138,54 @@ impl<'a> Toipe {
         let mut num_errors = 0;
         let mut num_chars_typed = 0;
 
-        let mut process_key = |key: Key| -> Result<bool, ToipeError> {
+        enum TestStatus {
+            // last key press did not quit/restart - more keys to be entered
+            NotDone,
+            // last letter was typed
+            Done,
+            // user wants to quit test
+            Quit,
+            // user wants to restart test
+            Restart,
+        }
+
+        impl TestStatus {
+            fn to_process_more_keys(&self) -> bool {
+                matches!(self, TestStatus::NotDone)
+            }
+
+            fn to_display_results(&self) -> bool {
+                matches!(self, TestStatus::Done)
+            }
+
+            fn to_restart(&self) -> bool {
+                matches!(self, TestStatus::Restart)
+            }
+        }
+
+        let mut process_key = |key: Key| -> Result<TestStatus, ToipeError> {
             match key {
-                Key::Ctrl(c) => match c {
-                    'c' => {
-                        return Ok(false);
+                Key::Ctrl('c') => {
+                    return Ok(TestStatus::Quit);
+                }
+                Key::Ctrl('r') => {
+                    return Ok(TestStatus::Restart);
+                }
+                Key::Ctrl('w') => {
+                    // delete last word
+                    while !matches!(input.last(), Some(' ') | None) {
+                        if input.pop().is_some() {
+                            self.tui.replace_text(
+                                Text::from(original_text[input.len()]).with_faint(),
+                            )?;
+                        }
                     }
-                    _ => {}
-                },
+                }
                 Key::Char(c) => {
                     input.push(c);
 
                     if input.len() >= original_text.len() {
-                        return Ok(false);
+                        return Ok(TestStatus::Done);
                     }
 
                     num_chars_typed += 1;
@@ -159,8 +205,7 @@ impl<'a> Toipe {
                     }
                 }
                 Key::Backspace => {
-                    let last_char = input.pop();
-                    if let Some(_) = last_char {
+                    if input.pop().is_some() {
                         self.tui
                             .replace_text(Text::from(original_text[input.len()]).with_faint())?;
                     }
@@ -170,7 +215,7 @@ impl<'a> Toipe {
 
             self.tui.flush()?;
 
-            Ok(true)
+            Ok(TestStatus::NotDone)
         };
 
         let mut keys = stdin.keys();
@@ -180,14 +225,12 @@ impl<'a> Toipe {
         // start the timer
         let started_at = Instant::now();
         // process first key
-        let res = process_key(key)?;
+        let mut status = process_key(key)?;
 
-        // process other keys if first key wasn't exit
-        if res {
+        if status.to_process_more_keys() {
             for key in &mut keys {
-                let res = process_key(key?)?;
-                // stop if key was exit (ctrl-c)
-                if !res {
+                status = process_key(key?)?;
+                if !status.to_process_more_keys() {
                     break;
                 }
             }
@@ -196,16 +239,35 @@ impl<'a> Toipe {
         // stop the timer
         let ended_at = Instant::now();
 
+        let (final_chars_typed_correctly, final_uncorrected_errors) =
+            input.iter().zip(original_text.iter()).fold(
+                (0, 0),
+                |(total_chars_typed_correctly, total_uncorrected_errors),
+                 (typed_char, orig_char)| {
+                    if typed_char == orig_char {
+                        (total_chars_typed_correctly + 1, total_uncorrected_errors)
+                    } else {
+                        (total_chars_typed_correctly, total_uncorrected_errors + 1)
+                    }
+                },
+            );
+
         let results = ToipeResults {
-            num_words: self.words.len(),
-            num_chars_typed,
-            num_chars_text: input.len(),
-            num_errors,
+            total_words: self.words.len(),
+            total_chars_typed: num_chars_typed,
+            total_chars_in_text: input.len(),
+            total_char_errors: num_errors,
+            final_chars_typed_correctly,
+            final_uncorrected_errors,
             started_at,
             ended_at,
         };
 
-        let to_restart = self.display_results(results.clone(), keys)?;
+        let to_restart = if status.to_display_results() {
+            self.display_results(results.clone(), keys)?
+        } else {
+            status.to_restart()
+        };
 
         Ok((to_restart, results))
     }
@@ -218,45 +280,44 @@ impl<'a> Toipe {
         self.tui.reset_screen()?;
 
         self.tui.display_lines::<&[Text], _>(&[
+            &[Text::from(format!(
+                "Took {}s for {} words of {}",
+                results.duration().as_secs(),
+                results.total_words,
+                self.config.text_name(),
+            ))],
             &[
                 Text::from(format!("Accuracy: {:.1}%", results.accuracy() * 100.0))
                     .with_color(color::Blue),
             ],
             &[Text::from(format!(
                 "Mistakes: {} out of {} characters",
-                results.num_errors, results.num_chars_text
-            ))],
-            &[Text::from(format!(
-                "Took {}s for {} words",
-                results.duration().as_secs(),
-                self.config.num_words,
+                results.total_char_errors, results.total_chars_in_text
             ))],
             &[
                 Text::from("Speed: "),
                 Text::from(format!("{:.1} wpm", results.wpm())).with_color(color::Green),
                 Text::from(" (words per minute)"),
             ],
-            &[],
-            &[
-                Text::from("Press "),
-                Text::from("r").with_color(color::Blue),
-                Text::from(" to restart, "),
-                Text::from("q").with_color(color::Blue),
-                Text::from(" to quit."),
-            ],
         ])?;
+        self.tui.display_lines_bottom(&[&[
+            Text::from("ctrl-r").with_color(color::Blue),
+            Text::from(" to restart, ").with_faint(),
+            Text::from("ctrl-c").with_color(color::Blue),
+            Text::from(" to quit ").with_faint(),
+        ]])?;
         // no cursor on results page
         self.tui.hide_cursor()?;
 
         // TODO: make this a bit more general
         // perhaps use a `known_keys_pressed` flag?
         let mut to_restart: Option<bool> = None;
-        while let None = to_restart {
+        while to_restart.is_none() {
             match keys.next().unwrap()? {
                 // press 'r' to restart
-                Key::Char('r') => to_restart = Some(true),
+                Key::Ctrl('r') => to_restart = Some(true),
                 // press 'q' to quit
-                Key::Char('q') => to_restart = Some(false),
+                Key::Ctrl('c') => to_restart = Some(false),
                 _ => {}
             }
         }
